// -*- c++ -*-

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

  SUBSYSTEM:  Decoder
  FILE:       AnimationStreamDec.h
  AUTHOR:     Mikael Bourges-Sevenier (France Telecom CNET/DIH/ATI)
  OVERVIEW:   Definition of AnimationStreamDecoder
  ---------------------------------------------------------------------------
  
    The purpose of the animation stream decoder is first, to decode the 
    animation mask sent at the beginning of an animation. It then allocates a 
    list of animated nodes. Second, for each animation frame, we call each 
    node of the list to decode itself.
    In addition, the decoder takes care of skipping frames for nodes willing 
    to do so.
    
-------------------------------------------------------------------------*/

#if !defined(_animationdec_h)	// Sentry, use file only if it's not already included.
#define _animationdec_h


#include "DecoderImp.h"
#include "ArithDec.h"
#define CPLUSPLUS

class MediaObject;

class AnimTime {
  int frame_rate; // the reference frame rate of the sequence
  int fr_seconds;   // the fractional reference frame rate
  bool fr_offset;    // true: the frame rate uses the NTSC frequency offset of 1000/1001
  long time_code;
  int tc_hours;  // 0-23
  int tc_minutes;// 0-59
  bool tc_mbit;   // marker bit always 1
  int tc_seconds;// 0-59
  int num_frames_skip;

  float fps; // frame per seconds

public:
  AnimTime() :
    fr_offset(FALSE),tc_mbit(TRUE)
  {
    setFrameRate(0);
    setTimeCode(0);
    setNumSkipFrames(0);
  }

  AnimTime(int frameRate, int timeCode, int framesSkip) :
    fr_offset(FALSE),tc_mbit(TRUE)
  {
    setFrameRate(frameRate);
    setTimeCode(timeCode);
    setNumSkipFrames(framesSkip);
  }

  bool hasFrameRate() { return (frame_rate != 0) ? true : false; }
  bool hasSkipFrames() { return (num_frames_skip != 0) ? true : false; }
  bool hasTimeCode() { return (time_code != 0) ? true : false; }

  void setFrameRate(int frameRate) { frame_rate = frameRate; }
  int getFrameRate() { return frame_rate; }
  int getFrameRateSeconds() { return fr_seconds; }
  void setFrameRateSeconds(int sec) { fr_seconds=sec; }
  bool getFrameRateOffset() { return fr_offset; }
  void setFrameRateOffset(bool b) { fr_offset=b; }

  void setNumSkipFrames(int framesSkip) { num_frames_skip = framesSkip; }
  int getNumSkipFrames() { return num_frames_skip; }
  void setTimeCode(long tc_millis) 
  { 
    int l= (int) (tc_millis / 1000);
    time_code=tc_millis/1000; 
    tc_hours = l / 3600;
    tc_minutes = l / 60;
    tc_seconds = l % 60;
  }
  
  long getTimeCode() { return time_code; }
  int getTimeCodeHours() { return tc_hours; }
  void setTimeCodeHours(int h) { tc_hours=h; }
  int getTimeCodeMinutes() { return tc_minutes; }
  void setTimeCodeMinutes(int m) { tc_minutes=m; }
  int getTimeCodeSeconds() { return tc_seconds; }
  void setTimeCodeSeconds(int s) { tc_seconds=s; }
  bool getTimeCodeMarkerBit() { return tc_mbit; }
  void setTimeCodeMarkerBit(bool mb) { tc_mbit=mb; }

  float getFPS() 
  {
    fps=frame_rate+fr_seconds/16.0;
    if (fr_offset) 
      fps *= (float) (1000.0/1001.0);
    return fps;
  }  
};

class AnimationStreamDecoder : public DecoderImp
{
private:
  virtual BOOL Setup (
		      const BYTE * pSpecificInfo,
		      int nSpecificInfoLength,
		      int nESNumber
		      );
  virtual bool ParseMask ();

  int readAFrame();
  int readFrameHeader();
  void readFrameData();
  bool checkAnimStartCode();
  float readFrameRate();
  void  readTimeCode();
  bool readSkipFrames();
  
public:
  virtual int GetOptimalOutputSize () {return 0; }
	virtual BOOL SetFormat (
		Capabilities &formatOptions,
		DecoderParams &decoderParams
	);
    
  virtual BOOL Decode (LPBYTE /*pInput*/, int /*nInputLength*/,
		       LPBYTE /*pOutput*/, int & /*rnOutputLength*/) { 
    return TRUE; 
  }

  //FTRD - begin - 7/10/2001 
  bool getRandomAccess() {return randomAccess;}
  void markNewAQP(bool state=true) {m_newAqp = state;}
  bool isNewAQP();
  //FTRD - end - 7/10/2001 
protected:
  // animation stream variables
  unsigned char *m_pMask; // the mask
  int maskSize; // and its size

  int nodeIDbits,routeIDbits; // BifsDecConfig variables
  

  /* The arithmetic decoder is handled by the decoder itself.
     Each animated field has is own model passed to the AAD.
   */
  ArithmeticDecoder aad;
	bool randomAccess;
  AnimTime m_aTime; // keep time stuff
  bool initialised; // do we have mask initialized ?
  int numNodes; // nodes in mask
  ZArray<bool> nodeMask; // active nodes switch
  ZArray<MediaObject *> pNodeMask; // ptrs to active nodes
  bool isIntra; // intra or predictive frame ?
  
  // bit manipulation variables
  unsigned char *bufptr,*startbuf;  // ptr to stream buffer
  unsigned char inbfr;    // input buffer
  unsigned int incnt;     // bit used counter in current read byte
  unsigned int inused;    // number of bits read
  unsigned int inbytecnt; // number of bytes read
  int bufSize;                     // input buffer unit size

  //FTRD - begin - 7/10/2001 
  FILE*		m_logfile;
  int		m_nbLogBits;
  bool		m_newAqp;
  //FTRD - end- 7/10/2001 
  
public:
  AnimationStreamDecoder ();
  //FTRD - begin - 7/10/2001 
  ~AnimationStreamDecoder ();
  //FTRD - end- 7/10/2001 

  virtual void Run();
  ArithmeticDecoder& getAAD() { return aad; }

  // bit manipulation methods
  void setInputBuffer(unsigned char *AUptr);
  unsigned char *getInputBuffer() { return bufptr; }
  void initBits();
  int readBits(int n);
  bool readBit();

 //FTRD - begin - 7/10/2001 
  int readBits(int n, char* szDesc, char* szComment=NULL,int logLevel=0);
  bool readBit(char* szDesc, char* szComment=NULL,int logLevel=0);
  float readFloat (char* szDesc, char* szComment=NULL,int logLevel=0);
  float readGenericFloat (bool useEfficientCoding, char* szDesc, char* szComment=NULL,int logLevel=0);
  void logEncoded(int val,char* szDesc, char* szComment=NULL,int logLevel=0);
  void log(int nValue, int nBits, char* szDesc, char* szComment,int logLevel=0);
  void log(char* szComment,int logLevel=0);
//FTRD - end- 7/10/2001 

  void alignBits();
  void alignInputBits();
  float readFloat ();
 //FTRD - begin - 7/10/2001 
  float readEfficientFloat (int* bitcount = NULL);
  float readGenericFloat (bool useEfficientCoding,int* bitcount = NULL);
  //FTRD - end - 7/10/2001 
  void adjustBits();
  void setBufSize (int inSize) { bufSize = inSize; }
  int getBitsIn() { return incnt; }
  void resetBitsIn() { incnt=0; inbfr=0; }
  int getCurrentBufSize() { return (bufSize - (int)inbytecnt); }
};

#endif                                      // _animationdec_h sentry.
